class Finder:
    """Interface defintion for a [finder](https://docs.python.org/3/glossary.html#term-finder).
    """
    
    def find_spec(self, fullname, path=None, target=None):
        raise NotImplementedError('Not Implemented Error: find_spec should be implemented')

    def get_loader(self):
        raise NotImplementedError('Not Implemented Error: get_loader should be implemented')




import os
import sys
import pytest
import threading
import time
from http.server import HTTPServer, SimpleHTTPRequestHandler
from pyimporter import enable_url_import, disable_url_import

class LiveServer:
    _stop = False
    thread = None
    value = 0
    httpd = None

    def __init__(self, path=None, address='127.0.0.1', port=8000):
        self.path = path
        self.address = address
        self.port = port
        self.ready = False

    def _start_server(self):
        address = (self.address, self.port)
        handler_class = SimpleHTTPRequestHandler
        if self.path is not None:
            os.chdir(self.path)
        with HTTPServer(address, handler_class) as httpd:
            self.httpd = httpd
            self.ready = True
            httpd.serve_forever()


    def start(self):
        if self.thread:
            raise Exception('Already running')
        self._stop = False
        self.thread = threading.Thread(name='live server',
                             target = self._start_server,
                             args = ())
        self.thread.setDaemon(True)
        self.thread.start()
        give_up_at_time = time.time() + 10
        while not self.ready:
            time.sleep(0.1)
            if time.time() > give_up_at_time:
                raise Exception("Live server didn't start on time. Giving up.")

    def stop(self):
        if self.thread:
            # self.thread._stop()
            self.httpd.shutdown()
            self.thread.join()
            self.thread = None
        pass

def pytest_addoption(parser):
    parser.addoption('--functional', action='store_true', dest="functional",
                 default=False, help="enable functional tests")

def pytest_configure(config):
    if not config.option.functional:
        setattr(config.option, 'markexpr', 'not functional')


@pytest.fixture(scope='session')
def fix_dir():
    dir = os.path.abspath(os.path.dirname(__file__)) + '/fixture'
    return dir

@pytest.fixture(scope='session')
def live_server(fix_dir):
    server = LiveServer(path=fix_dir)
    server.start()
    yield server
    server.stop()

@pytest.fixture
def urlimport(live_server):
    enable_url_import()
    sys.path.append('http://127.0.0.1:8000/')
    yield None
    sys.path.remove('http://127.0.0.1:8000/')
    disable_url_import()
    sys.path_importer_cache.clear()





class Loader:
    """Interface defintion for a [Loader](https://docs.python.org/3/glossary.html#term-loader)
    See also [Loader reference](https://docs.python.org/3/reference/import.html#loaders).
    """
    def create_module(self, spec):
        raise NotImplementedError('Not Implemented Error: create_module should be implemented')

    def exec_module(self, mod):
        raise NotImplementedError('Not Implemented Error: exec_module should be implemented')

class BaseLoader(Loader):
    def create_module(self, spec):
        return None

    def get_source_path(self):
        raise NotImplementedError('Not Implemented Error: get_source should be implemented')

    def get_source(self):
        raise NotImplementedError('Not Implemented Error: get_source should be implemented')

    def exec_module(self, mod):
        mod.__path__ = self.get_source_path()
        exec(self.get_source(), mod.__dict__)
    


import sys
import os
import logging
from urllib.request import urlopen
from urllib.parse import urljoin
from urllib.error import HTTPError
import importlib.util
from functools import wraps
from .finder import Finder
from .loader import BaseLoader

class PathHook:
    """Abstract [path hook](https://docs.python.org/3/library/sys.html#sys.path_hooks).
    """
    def get_finder(self):
        raise NotImplementedError('Not Implemented Error: get_finder should be implemented')

    def __call__(self, path):
        self.base_path = path
        return self.get_finder()



class PathImporter(PathHook, BaseLoader, Finder):

    @wraps(Finder.find_spec)
    def find_spec(self, fullname, path=None, target=None):
        self.spec = None
        self.fullname = fullname
        for is_package in [False, True]:
            try:
                self.is_package = is_package
                self.source = self.read_source()
                self.create_spec()
                break
            except FileNotFoundError:
                pass
        return self.spec

    @wraps(PathHook.get_finder)
    def get_finder(self):
        return self

    @wraps(Finder.get_loader)
    def get_loader(self):
        return self

    @wraps(BaseLoader.get_source)
    def get_source(self):
        return self.source
    
    @wraps(BaseLoader.get_source_path)
    def get_source_path(self):
        return [self.get_location()]

    def create_spec(self):
        self.spec = importlib.util.spec_from_loader(
                    self.fullname, 
                    self.get_loader(), 
                    origin=self.get_fullpath(), 
                    is_package=self.is_package)
        return self.spec

    def path_join(self, base_path, path):
        return os.path.join(base_path, path)

    def get_location(self):
        return self.path_join(self.base_path, self.fullname.replace('.', '/'))

    def get_fullpath(self):
        suffix = '/__init__.py' if self.is_package else '.py'
        fullpath = self.get_location() + suffix
        return fullpath

    def read_source(self):
        fullpath = self.get_fullpath()
        source = None
        with open(fullpath) as file:
            source = file.read().decode().replace("\r\n", "\n")
        return source

class UrlImporter(PathImporter):

    @wraps(PathImporter.path_join)
    def path_join(self, base_path, path):
        return urljoin(base_path, path)

    @wraps(PathImporter.read_source)
    def read_source(self):
        fullpath = self.get_fullpath()
        source = None
        try:
            logging.debug(f'urlopen: {fullpath}')
            with urlopen(fullpath) as file:
                source = file.read().decode().replace("\r\n", "\n")
        except HTTPError as exc:
            if '404' in str(exc):
                raise FileNotFoundError(f'File Not Found Error: {self.fullname} ({fullpath})')
            raise
        return source





from .common import register_finder, enable_url_import, disable_url_import
from .importer import PathImporter, UrlImporter

__all__ = ['PathImporter', 'UrlImporter', 'register_finder', 'enable_url_import', 'disable_url_import']



import sys
from .importer import UrlImporter

_url_importer = None

def register_finder(finder, as_priority=False):
    global _url_importer
    if finder in sys.path_hooks:
        sys.path_hooks.remove(finder)
    if as_priority:
        sys.path_hooks.insert(0, finder)
    else:
        sys.path_hooks.append(finder)

def disable_url_import():
    global _url_importer
    if _url_importer in sys.path_hooks:
        sys.path_hooks.remove(_url_importer)

def enable_url_import(as_priority=False):
    global _url_importer
    if _url_importer is None:
        _url_importer = UrlImporter()
    register_finder(_url_importer, as_priority)



import time
import pytest
from pyimporter import enable_url_import

@pytest.mark.functional
def test_Should_HavePackageImported_When_ImportFromUrl(urlimport):
    import agent
    assert '007' == agent.name 

@pytest.mark.functional
def test_Should_HaveModuleImported_When_ImportFromUrl(urlimport):
    from agent import actions
    assert 'Baff' == actions.bark() 

@pytest.mark.functional
def test_Should_RiseExecption_When_ImportedCodeFailsToCompile(urlimport):
    with pytest.raises(IndentationError):
        from agent import bad_syntax_code

@pytest.mark.functional
def test_Should_RiseExecption_When_ImportedCodeFailsToRun(urlimport):
    with pytest.raises(NameError):
        from agent import bad_execution_code

